!
!  Bookkeeping with Region partitioning algorithm
!  New algorithm
!
      subroutine bookkeep_rp
      use constants
      use variables
      use bookkeep_param
      use rp_param
      use debug_variables
      implicit double precision (a-h,o-z)
      double precision :: sdx,sdy,sdz
      integer :: istat
      istat=0

      atime=dtime(btime)
      if (.not. allocated (knum)) then
         allocate (knum(n))
      endif

!     Region partitioning
      ! Check if region partitioning is available (everytime cuz cell size may vary)
      if      (hmat(1,1)<frc*1.1) then
         stop 'Region partitioning failure: Lx is too small'
      else if (hmat(2,2)<frc*1.1) then
         stop 'Region partitioning failure: Ly is too small'
      else if (hmat(3,3)<frc*1.1) then
         stop 'Region partitioning failure: Lz is too small'
      endif
      if (abs(hmat(1,2)/hmat(1,1))>0.1) stop 'Partitioning is available only for orthorhombic cell'
      if (abs(hmat(1,3)/hmat(1,1))>0.1) stop 'Partitioning is available only for orthorhombic cell'
      if (abs(hmat(2,1)/hmat(2,2))>0.1) stop 'Partitioning is available only for orthorhombic cell'
      if (abs(hmat(2,3)/hmat(2,2))>0.1) stop 'Partitioning is available only for orthorhombic cell'
      if (abs(hmat(3,1)/hmat(3,3))>0.1) stop 'Partitioning is available only for orthorhombic cell'
      if (abs(hmat(3,2)/hmat(3,3))>0.1) stop 'Partitioning is available only for orthorhombic cell'
      
! Offset: needed for cell w/o pbc
      offx=1.0d10; offy=1.0d10; offz=1.0d10
      xmax=-1.0d10;ymax=-1.0d10;zmax=-1.0d10
      do i=1,n
         if (offx>rx(i)) offx=rx(i)
         if (offy>ry(i)) offy=ry(i)
         if (offz>rz(i)) offz=rz(i)
         if (xmax<rx(i)) xmax=rx(i)
         if (ymax<ry(i)) ymax=ry(i)
         if (zmax<rz(i)) zmax=rz(i)
      enddo
      if (pbcx) offx=0.0d0
      if (pbcy) offy=0.0d0
      if (pbcz) offz=0.0d0

      ! Set grids for partitioning (only once)
      ! npc{x,y,z}: number of partitioning grids
      ! spc{x,y,z}: size of partitioning lattices
      ! nnor: # of partition cells
      ! nnir: max. # of atoms in each region
      ! idpart: cell number where atom is sitting
!!!      if (.not. allocated (nir)) then
      if (allocated(nir)) then
         deallocate(nir,ncl,coor,idpart)
      endif
!         npcx=hmat(1,1)/frc;  npcy=hmat(2,2)/frc;  npcz=hmat(3,3)/frc
         spcx=hmat(1,1)/npcx; spcy=hmat(2,2)/npcy; spcz=hmat(3,3)/npcz
         if (.not.pbcx) spcx=(xmax-offx)/npcx
         if (.not.pbcy) spcy=(ymax-offy)/npcy
         if (.not.pbcz) spcz=(zmax-offz)/npcz
!         if (npcx < 3) stop 'please set npcx > 2'
!         if (npcy < 3) stop 'please set npcy > 2'
!         if (npcz < 3) stop 'please set npcz > 2'
         if (spcx < frc) stop 'npcx too large'
         if (spcy < frc) stop 'npcy too large'
         if (spcz < frc) stop 'npcz too large'
         nnor=npcx*npcy*npcz
         nnir=n/nnor+200
         allocate(nir(nnor,nnir),ncl(nnor,26,4),coor(nnor,3),idpart(n))
!!!      endif
!     atom mapping to partitioning cells
      
      nir=0; ncl=0; coor=0.0d0; idpart=0
      icc=0
      icountall=0
! Offset: needed for cell w/o pbc
!      offx=1.0d10; offy=1.0d10; offz=1.0d10
!      do i=1,n
!         if (offx>rx(i)) offx=rx(i)
!         if (offy>ry(i)) offy=ry(i)
!         if (offz>rz(i)) offz=rz(i)
!      enddo
!      if (pbcx) offx=0.0d0
!      if (pbcy) offy=0.0d0
!      if (pbcz) offz=0.0d0
!
      do ix=1,npcx
         xx=spcx*(ix-1)+offx
         do iy=1,npcy
            yy=spcy*(iy-1)+offy
            do iz=1,npcz
               zz=spcz*(iz-1)+offz
               icc=icc+1
               coor(icc,1)=xx; coor(icc,2)=yy; coor(icc,3)=zz
               icount=0
               do i=1,n
                  if  ((rx(i)>=xx-1.0d-11).and.(rx(i)<=xx+spcx+1.0d-11)&
                  .and.(ry(i)>=yy-1.0d-11).and.(ry(i)<=yy+spcy+1.0d-11)&
                  .and.(rz(i)>=zz-1.0d-11).and.(rz(i)<=zz+spcz+1.0d-11)) then
                     if (idpart(i)/=0) cycle
                     idpart(i)=icc
                     icount=icount+1
                     icountall=icountall+1
                     nir(icc,icount)=i
                     if (icount>nnir) then
                     print *,'ERROR (bookkeep_rp.F)'
                     print *,'number of atoms'
                     print *,'in one partition over limit'
                     stop 
                     endif
                  endif
               enddo
            enddo
         enddo
      enddo
!
!      print *,'CELL:',hmat(1,1)/ang,hmat(2,2)/ang,hmat(3,3)/ang
!      print *,'OFF',offx/ang,offy/ang,offz/ang
!      print *,'MAX',xmax/ang,ymax/ang,zmax/ang
!      print *,'SPC',spcx/ang,spcy/ang,spcz/ang
      do i=1,n
         if (idpart(i)==0) then
            print *,i,'-th atom is not assigned to any partition.'
            print *,'x,y,z(ang):',rx(i)/ang,ry(i)/ang,rz(i)/ang
         endif
      enddo
!
      if (icountall/=n) then
         print *,'ERROR (bookkeep_rp.F)'
         print *,'number of atoms mapped:',icountall
         stop
      endif
!      do i=1,nnor
!         print *,i,real(1.0e10*coor(i,1)),real(1.0e10*coor(i,2)),real(1.0e10*coor(i,3))
!      enddo
!     prepare list of neighboring partitions
      do ii=1,nnor
         icount=0
         do jj=1,nnor
            if (ii==jj) cycle
            xx=coor(jj,1)-coor(ii,1); yy=coor(jj,2)-coor(ii,2); zz=coor(jj,3)-coor(ii,3)
            irepx=0; irepy=0; irepz=0
            if (pbcx) then
               if      (xx >  spcx*1.1) then
                  irepx = -1; xx = xx - hmat(1,1)
               else if (xx < -spcx*1.1) then
                  irepx =  1; xx = xx + hmat(1,1)
               endif
            endif
            if (pbcy) then
               if      (yy >  spcy*1.1) then
                  irepy = -1; yy = yy - hmat(2,2)
               else if (yy < -spcy*1.1) then
                  irepy =  1; yy = yy + hmat(2,2)
               endif
            endif
            if (pbcz) then
               if      (zz >  spcz*1.1) then
                  irepz = -1; zz = zz - hmat(3,3)
               else if (zz < -spcz*1.1) then
                  irepz =  1; zz = zz + hmat(3,3)
               endif
            endif
            if ((abs(xx) <= spcx*1.1).and.(abs(yy) <= spcy*1.1).and.(abs(zz) <= spcz*1.1)) then
               icount=icount+1
               if (icount.gt.26) then
                  stop 'Neighbor cell error: More than 26 are found'
               endif
               ncl(ii,icount,1)=jj; ncl(ii,icount,2)=irepx; ncl(ii,icount,3)=irepy; ncl(ii,icount,4)=irepz
            endif
         enddo
      enddo
!     Result of region partitioning
!      print *,'Number of cells'
!      print *,nnor, '(',npcx, npcy, npcz,')'
!      print *,'Cell size'
!      print *,real(1.0e10*spcx),real(1.0e10*spcy),real(1.0e10*spcz)
!      print *,'Cell coordinates'
!      do i=1,nnor
!         print 910,i,real(1.0e10*coor(i,1)),real(1.0e10*coor(i,2)),real(1.0e10*coor(i,3))
!      enddo
!      print *,'Neighbor list'
!      do i=1,nnor
!         print *,'List of',i,'-th partition'
!         do j=1,26
!            if (ncl(i,j,1) == 0) cycle
!            print 900,ncl(i,j,1),ncl(i,j,2),ncl(i,j,3),ncl(i,j,4) &
!            ,real(1.0e10*coor(ncl(i,j,1),1)),real(1.0e10*coor(ncl(i,j,1),2)),real(1.0e10*coor(ncl(i,j,1),3))
!         enddo
!      enddo
!      stop
 900  format (4(i4,1x),3(f10.5,1x))
 910  format (1(i4,1x),3(f10.5,1x))


!      vave=sqrt(3.0d0*bl*temp/wm)
!      frc=rc+vave*2.0d0*dt*dble(nbk)*5.0d0
!      print *,'RC=',rc/ang
!      print *,'FRC=',frc/ang
      frc2=frc**2

! Check cell size, determine # of replica cells to be serched
      nrpx=0; nrpy=0; nrpz=0
!      if (pbcx) nrpx=int(frc*2.0/hmat(1,1))+1
!      if (pbcy) nrpy=int(frc*2.0/hmat(2,2))+1
!      if (pbcz) nrpz=int(frc*2.0/hmat(3,3))+1
      if (pbcx) nrpx=int(frc*2.0/sqrt(hmat(1,1)**2+hmat(2,1)**2+hmat(3,1)**2))+1
      if (pbcy) nrpy=int(frc*2.0/sqrt(hmat(1,2)**2+hmat(2,2)**2+hmat(3,2)**2))+1
      if (pbcz) nrpz=int(frc*2.0/sqrt(hmat(1,3)**2+hmat(2,3)**2+hmat(3,3)**2))+1
!      print *,'Replica cells searched:',nrpx,nrpy,nrpz
      if (n>1000) then
         if ((nrpx>1).or.(nrpy>1).or.(nrpz>1)) then
            iwarn=iwarn+1
            print *,'WARNING:'
            print *,'The cell is large and is short in one direction,'
            print *,'which cannot be solved efficiently.'
            print *,'please duplicate the cell in the direction.'
         endif
      endif
      if ((.not.allocated (idpt)).and.(nbook>0)) then
         allocate (idpt(nbook,4))
      endif
      if (bookkeep_algo /= 1) go to 5000
!!! Normal algorithm
!!! Count and allocate list vector
 1000 continue
      if (.not. allocated (idpt)) then
         kns=0
         do k=1,n
            kn=0
            sxk=rx(k)
            syk=ry(k)
            szk=rz(k)
            do j=k,n
               if ((nrpx<=1).and.(nrpy<=1).and.(nrpz<=1)) then
                  sdx=sxk-rx(j)
                  sdy=syk-ry(j)
                  sdz=szk-rz(j)
                  ix=0; iy=0; iz=0
                  call search_neighbor(sdx,sdy,sdz,frc,ix,iy,iz,istat)
                  if (istat == 1) then
                     r2=sdx**2+sdy**2+sdz**2
                     if (r2 < frc2) then
                        kn=kn+1
                        kns=kns+1
                     endif
                  endif
               else
                  do ix=-nrpx,nrpx; do iy=-nrpy,nrpy; do iz=-nrpz,nrpz
                     sdx=sxk-rx(j)
                     sdy=syk-ry(j)
                     sdz=szk-rz(j)
                     call calc_replica(sdx,sdy,sdz,frc,ix,iy,iz,istat)
                     if (istat == 1) then
                        r2=sdx**2+sdy**2+sdz**2
                        if (r2 < frc2) then
                           kn=kn+1
                           kns=kns+1
                        endif
                     endif
                  enddo; enddo; enddo
               endif
            enddo
            knum(k)=kn
         enddo
         nbook=kns*1.5
         print *,'nbook should be >',kns
         print *,'nbook is set to',nbook
         allocate (idpt(nbook,4))
      endif
      atime=dtime(btime)
      if (btime(2)>1) print *,'Bookkeep counting:',btime,'  sec'
!!! Count end
      kns=0
      do 3200 k=1,n
         kn=0
         sxk=rx(k)
         syk=ry(k)
         szk=rz(k)
!         do 3300 j=k,n
         idp=idpart(k)
         do 3300 inc=0,26
            if (inc==0) then
               idpnei=idp
            else
               idpnei=ncl(idp,inc,1)
            endif
            if (idpnei==0) cycle
            do 3301 jcount=1,nnir
               j=nir(idpnei,jcount)
               if (j==0) cycle
               if (j<k) cycle

            if ((nrpx<=1).and.(nrpy<=1).and.(nrpz<=1)) then
               sdx=sxk-rx(j)
               sdy=syk-ry(j)
               sdz=szk-rz(j)
               call search_neighbor(sdx,sdy,sdz,frc,ix,iy,iz,istat)
               if (istat == 1) then
                  r2=sdx**2+sdy**2+sdz**2
                  if (r2 < frc2) then
                     kn=kn+1
                     kns=kns+1
                     if (kns <= nbook) then
                        idpt(kns,1)=j
                        idpt(kns,2)=ix
                        idpt(kns,3)=iy
                        idpt(kns,4)=iz
                     else
                        print *,'WARNING:'
                        iwarn=iwarn+1
                        print *,'nbook should be larger. expanded.'
                        print *,'Pls set larger nbook to save time'
                        deallocate (idpt)
                        go to 1000
                     endif
                  endif
               endif
            else
               do ix=-nrpx,nrpx; do iy=-nrpy,nrpy; do iz=-nrpz,nrpz
                  sdx=sxk-rx(j)
                  sdy=syk-ry(j)
                  sdz=szk-rz(j)
                  call calc_replica(sdx,sdy,sdz,frc,ix,iy,iz,istat)
                  if (istat == 1) then
                     r2=sdx**2+sdy**2+sdz**2
                     if (r2 < frc2) then
                        kn=kn+1
                        kns=kns+1
                        if (kns <= nbook) then
                           idpt(kns,1)=j
                           idpt(kns,2)=ix
                           idpt(kns,3)=iy
                           idpt(kns,4)=iz
                        else
                           print *,'WARNING:'
                           iwarn=iwarn+1
                           print *,'nbook should be larger. expanded.'
                           print *,'Pls set larger nbook to save time'
                           deallocate (idpt)
                           go to 1000
                        endif
                     endif
                  endif
               enddo; enddo; enddo
            endif
 3301       continue
 3300    continue
         knum(k)=kn
 3200 continue
      if (ilog == 1) then
         print *,'NUMBER OF BOOKKEEPING : ',kns
      endif
!      if (kns.gt.nbook) then
!         print *,'!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
!         print *,'!!! ERROR in BOOKKEEPING !!!'              
!         print *,'!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
!      endif

      atime=dtime(btime)
      print *,'Bookkeep+RP (NORMAL algorithm)',btime,'  sec'
      return

!!! Double algorithm
!!! Count and allocate list vector
 5000 continue
      if (.not. allocated (idpt)) then
         kns=0
         do k=1,n
            kn=0
            sxk=rx(k)
            syk=ry(k)
            szk=rz(k)
            do j=1,n
               if ((nrpx<=1).and.(nrpy<=1).and.(nrpz<=1)) then
                  sdx=sxk-rx(j)
                  sdy=syk-ry(j)
                  sdz=szk-rz(j)
                  ix=0; iy=0; iz=0
                  call search_neighbor(sdx,sdy,sdz,frc,ix,iy,iz,istat)
                  if (istat == 1) then
                     r2=sdx**2+sdy**2+sdz**2
                     if (r2 < frc2) then
                        kn=kn+1
                        kns=kns+1
                     endif
                  endif
               else
                  do ix=-nrpx,nrpx; do iy=-nrpy,nrpy; do iz=-nrpz,nrpz
                     sdx=sxk-rx(j)
                     sdy=syk-ry(j)
                     sdz=szk-rz(j)
                     call calc_replica(sdx,sdy,sdz,frc,ix,iy,iz,istat)
                     if (istat == 1) then
                        r2=sdx**2+sdy**2+sdz**2
                        if (r2 < frc2) then
                           kn=kn+1
                           kns=kns+1
                        endif
                     endif
                  enddo; enddo; enddo
               endif
            enddo
            knum(k)=kn
         enddo
         nbook=kns*1.5
         allocate (idpt(nbook,4))
      endif
!!! Count end

      kns=0
      do k=1,n
         kn=0
         sxk=rx(k)
         syk=ry(k)
         szk=rz(k)
!         do j=1,n
         idp=idpart(k)
         do 5300 inc=0,26
            if (inc==0) then
               idpnei=idp
            else
               idpnei=ncl(idp,inc,1)
            endif
            if (idpnei==0) cycle
            do 5301 jcount=1,nnir
               j=nir(idpnei,jcount)
               if (j==0) cycle
!               if (j<k) cycle

            if ((nrpx<=1).and.(nrpy<=1).and.(nrpz<=1)) then
               sdx=sxk-rx(j)
               sdy=syk-ry(j)
               sdz=szk-rz(j)
               call search_neighbor(sdx,sdy,sdz,frc,ix,iy,iz,istat)
               if (istat == 1) then
                  r2=sdx**2+sdy**2+sdz**2
                  if (r2 < frc2) then
                     kn=kn+1
                     kns=kns+1
                     if (kns <= nbook) then
                        idpt(kns,1)=j
                        idpt(kns,2)=ix
                        idpt(kns,3)=iy
                        idpt(kns,4)=iz
                     else
                        print *,'WARNING:'
                        iwarn=iwarn+1
                        print *,'nbook should be larger. expanded.'
                        print *,'Pls set larger nbook to save time'
                        deallocate (idpt)
                        go to 5000
                     endif
                  endif
               endif
            else
               do ix=-nrpx,nrpx; do iy=-nrpy,nrpy; do iz=-nrpz,nrpz
                  sdx=sxk-rx(j)
                  sdy=syk-ry(j)
                  sdz=szk-rz(j)
                  call calc_replica(sdx,sdy,sdz,frc,ix,iy,iz,istat)
                  if (istat == 1) then
                     r2=sdx**2+sdy**2+sdz**2
                     if (r2 < frc2) then
                        kn=kn+1
                        kns=kns+1
                        if (kns <= nbook) then
                           idpt(kns,1)=j
                           idpt(kns,2)=ix
                           idpt(kns,3)=iy
                           idpt(kns,4)=iz
                        else
                           print *,'WARNING:'
                           iwarn=iwarn+1
                           print *,'nbook should be larger. expanded.'
                           print *,'Pls set larger nbook to save time'
                           deallocate (idpt)
                           go to 5000
                        endif
                     endif
                  endif
               enddo; enddo; enddo
            endif
!         end do  !
!         end do
 5301       continue
 5300    continue
         knum(k)=kn
      end do
      if (ilog == 1) then
         print *,'NUMBER OF BOOKKEEPING : ',kns
      endif
!      if (kns.gt.nbook) then
!         print *,'!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
!         print *,'!!! ERROR in BOOKKEEPING !!!'              
!         print *,'!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
!      endif

      atime=dtime(btime)
      print *,'Bookkeep+RP (DOUBLE algorithm)',btime,'  sec'
      return

      end


